from urllib.request import urlretrieve
from pydantic import validate_call

class BaseStaticDownloader:
    url: str

    @validate_call
    def download(self, output: str):
        urlretrieve(self.url, output)


from pydantic import validate_call

from .config import ELEMENTS_CONFIG

@validate_call
def calculate_strongs(elements: list[str]) -> list[str]:
    strongs = []

    for element in elements:
        preview_strongs = ELEMENTS_CONFIG[element]["strongs"]

        for element in preview_strongs:
            if not element in strongs:
                strongs.append(element)

    return strongs

@validate_call
def calculate_weaknesses(first_element: str) -> list[str]:
    weaknesses = []

    for element in ELEMENTS_CONFIG.keys():
        element_strongs = ELEMENTS_CONFIG[element]["strongs"]

        if first_element in element_strongs:
            weaknesses.append(element)

    return weaknesses


from typing import Optional

from .config import get_music_name_from_tag
from ..base import BaseStaticDownloader

class GeneralMusic(BaseStaticDownloader):
    def __init__(
        self,
        music_name: Optional[str] = None,
        tag: Optional[str] = None
    ) -> None:
        if not tag and not music_name:
            raise ValueError("You must enter at least one name or tag!")

        if tag:
            music_name = get_music_name_from_tag(tag)

        if not music_name:
            raise ValueError("Please enter a valid song name! If you have entered a tag please check it as it was not possible to get a song name from it.")

        self.url = f"http://dci-static-s1.socialpointgames.com/static/dragoncity/mobile/sounds/music/{music_name}.mp3"


from pydantic import validate_call
from typing import Optional, Any
import json
import gzip

@validate_call
def read_file(file_path: str, encoding: str = "utf-8") -> str:
    with open(file_path, "r", encoding = encoding) as file:
        content = file.read()

    return content

@validate_call
def read_compressed_file(file_path: str, encoding: str = "utf-8") -> str:
    with gzip.open(file_path, "rt", encoding=encoding) as file:
        content = file.read()

    return content

@validate_call
def write_file(file_path: str, content: str, encoding: str = "utf-8") -> None:
    with open(file_path, "w", encoding = encoding) as file:
        file.write(content)

@validate_call
def write_compressed_file(file_path: str, content: str, encoding: str = "utf-8") -> None:
    with gzip.open(file_path, "wt", encoding = encoding) as file:
        file.write(content)

@validate_call
def read_json_file(file_path: str, encoding: str = "utf-8") -> Any:
    data_string = read_file(file_path, encoding = encoding)
    data = json.loads(data_string)
    return data

@validate_call
def read_compressed_json_file(file_path: str, encoding: str = "utf-8") -> Any:
    data_string = read_compressed_file(file_path, encoding = encoding)
    data = json.loads(data_string)
    return data

@validate_call
def write_json_file(
    file_path: str,
    data: Any,
    indent: Optional[int] = None,
    encoding: str = "utf-8"
) -> None:
    data_string = json.dumps(data, indent = indent)
    write_file(file_path, data_string, encoding)

@validate_call
def write_compressed_json_file(
    file_path: str,
    data: Any,
    indent: Optional[int] = None,
    encoding: str = "utf-8"
) -> None:
    data_string = json.dumps(data, indent = indent)
    write_compressed_file(file_path, data_string, encoding)


from urllib.request import urlretrieve, build_opener, install_opener
from pydantic import validate_call
from zipfile import ZipFile
import tensorflow as tf
import numpy as np
import pathutil
import logging
import os

@validate_call
def load_labels(file_path: str) -> list[str]:
    with open(file_path, "r") as file:
        lables = [ line.strip() for line in file.readlines() ]
        
    return lables

class BaseAIConfig:
    @validate_call
    def __init__(
        self,
        model_zip_url: str,
        model_out_dir: str,
        model_zip_filename: str,
        model_filename: str,
        labels_filename: str
    ) -> None:
        self.model_zip_url = model_zip_url
        self.model_out_dir = model_out_dir
        self.model_zip_filename = model_zip_filename
        self.model_filename = model_filename
        self.labels_filename = labels_filename

    @property
    def model_zip_file_path(self) -> str:
        return os.path.join(self.model_out_dir, self.model_zip_filename)

    @property
    def model_file_path(self) -> str:
        return os.path.join(self.model_out_dir, self.model_filename)

    @property
    def labels_file_path(self) -> str:
        return os.path.join(self.model_out_dir, self.labels_filename)

class BaseAI:
    @validate_call(config=dict(arbitrary_types_allowed=True))
    def __init__(self, config: BaseAIConfig) -> None:
        self.config = config

        if not os.path.exists(config.model_file_path) or not os.path.exists(config.labels_file_path):
            self._download_model_zip()
            self._unzip_model_zip()
            self._delete_model_zip()

        self._load_model()

    def _load_model(self):
        logging.info(f"Loading '{self.__class__.__name__}' ai model")

        np.set_printoptions(suppress=True)

        self.model = tf.keras.models.load_model(self.config.model_file_path)
        self.model.labels = load_labels(self.config.labels_file_path)

        self.data = np.ndarray(shape=(1, 224, 224, 3), dtype=np.float32)

    def _download_model_zip(self) -> None:
        logging.info(f"Downloading model AI from: {self.config.model_zip_url}")

        if not os.path.exists(self.config.model_out_dir):
            pathutil.mktree(self.config.model_out_dir)

        opener = build_opener()
        opener.addheaders = [("User-Agent", "Mozilla/5.0")]
        install_opener(opener)

        urlretrieve(self.config.model_zip_url, self.config.model_zip_file_path)

    def _unzip_model_zip(self) -> None:
        logging.info(f"Extracting model from: {self.config.model_zip_file_path}")

        with ZipFile(self.config.model_zip_file_path, "r") as file:
            file.extractall(self.config.model_out_dir)

    def _delete_model_zip(self) -> None:
        logging.info(f"Deleting model zip from: {self.config.model_zip_file_path}")
        os.remove(self.config.model_zip_file_path)

    def _delete_model_file(self) -> None:
        logging.info(f"Deleting model from: {self.config.model_file_path}")
        os.remove(self.config.model_file_path)

    def _delete_labels_file(self) -> None:
        logging.info(f"Deleting labels from: {self.config.labels_file_path}")
        os.remove(self.config.labels_file_path)

    def update(self) -> None:
        logging.info(f"Updating the model: {self.__class__.__name__}")
        
        self._delete_model_file()
        self._delete_labels_file()
        self._download_model_zip()
        self._unzip_model_zip()
        self._delete_model_zip()
        self._load_model()


from pydantic import validate_call
from PIL import Image, ImageOps
import numpy as np

from .base import BaseAI, BaseAIConfig

class PhaseDetectorAI(BaseAI):
    def __init__(self) -> None:
        config = BaseAIConfig(
            model_zip_url = "https://cdn.discordapp.com/attachments/1057469643402526781/1099397032805933148/phase_detector_model.zip",
            model_out_dir = ".dcutils/ai_data/phase_detector/",
            model_zip_filename = "phase_detector_model.zip",
            model_filename = "keras_model.h5",
            labels_filename = "labels.txt"
        )

        super().__init__(config)
    
    @validate_call
    def detect(self, image_path: str) -> dict:
        image_size = (224, 224)

        image = Image.open(image_path).convert("RGB")
        image = ImageOps.fit(
            image,
            image_size,
            Image.Resampling.LANCZOS
        )

        image_array = np.asarray(image)
        normalized_image_array = (image_array.astype(np.float32) / 127.5) - 1

        self.data[0] = normalized_image_array

        prediction = self.model.predict(self.data)

        top_index = np.argmax(prediction)

        result = {
            "phase": self.model.labels[top_index][2:].strip().lower(),
            "confidence_score": float(prediction[0][top_index])
        }

        return result


from typing import Optional, Union, Any, Self
from pydantic import validate_call
from pyfilter import DictListFilter
import httpx
import json

from ..._utils.file_manager import (
    write_json_file,
    read_json_file,
    write_compressed_json_file,
    read_compressed_json_file
)

LocalizationDict = dict[str, str]
LocalizationList = list[LocalizationDict]

class Localization:
    _list: list
    _dict: dict

    @validate_call
    def __init__(
        self,
        language: Optional[str] = None,
        loc: Union[list[dict], dict, None] = None
    ) -> None:
        if language:
            self._list: LocalizationList = self.fetch(language)
            self._dict: LocalizationDict = DictListFilter(self._list).merge_dicts()

        elif loc:
            self._load(loc)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({json.dumps(self.dict, indent=4)[:300]}" + "...})"

    def __str__(self) -> str:
        return self.__repr__()

    # >> Desativados temporariamente por conta de erros
    # @validate_call
    # def __getitem__(self, key: str) -> str:
    #     result = self.get_value_from_key(key)

    #     if result is None:
    #         raise KeyError(key)

    #     return result

    # @validate_call
    # def __getattr__(self, name: str) -> str:
    #     result = self.get_value_from_key(name)

    #     if result is None:
    #         raise AttributeError(name)

    #     return result
    # >>

    @classmethod
    @validate_call
    def load_file(cls, file_path: str, encoding: str = "utf-8") -> Self:
        loc = read_json_file(file_path, encoding = encoding)
        return cls(loc = loc)

    @classmethod
    @validate_call
    def load(cls, loc: Union[list, dict]) -> Self:
        loc_obj = cls()
        object_type = type(loc)

        if object_type == list:
            loc_obj._list = loc
            loc_obj._dict = DictListFilter(loc).merge_dicts()

        elif object_type == dict:
            loc_obj._dict = loc
            loc_obj._list = []

            for key, value in loc.items():
                dict_ = { key: value }
                loc_obj._list.append(dict_)

        else:
            raise ValueError(f"'{object_type}' is an invalid type to load a localization!")

        return loc_obj

    @validate_call
    def save_file(
        self,
        file_path: str,
        from_: str = "dict"
    ) -> None:
        if from_ == "dict":
            data = self._dict

        elif from_ == "list":
            data = self._list

        else:
            ValueError()

        write_json_file(file_path, data)

    @validate_call
    def save_compressed_file(
        self,
        file_path: str,
        encoding: str = "utf-8",
        indent: Optional[int] = None
    ) -> None:
        write_compressed_json_file(file_path, self.dict, indent, encoding)

    @classmethod
    @validate_call
    def load_compressed_file(
        cls,
        file_path: str,
        encoding: str = "utf-8"
    ) -> Self:
        loc = read_compressed_json_file(file_path, encoding)
        return cls(loc = loc)

    @classmethod
    def fetch(cls, language: str) -> list[dict[str, str]]:
        endpoint_url = f"https://sp-translations.socialpointgames.com/deploy/dc/android/prod/dc_android_{language}_prod_wetd46pWuR8J5CmS.json"

        response = httpx.get(endpoint_url)
        data = response.json()
        return data

    @validate_call
    def _load(self, loc: Union[list, dict]):
        type_ = type(loc)

        if type_ == list:
            self._load_list(loc)

        elif type_ == dict:
            self._load_dict(loc)

        else:
            raise ValueError(f"{type_} is an invalid type to load a localization")

    @validate_call
    def _load_list(self, loc: list[dict]):
        self._list = loc
        self._dict = DictListFilter(loc).merge_dicts()

    @validate_call
    def _load_dict(self, loc: dict):
        self._dict = loc
        self._list = []

        for key, value in loc.items():
            dict_ = { key: value }
            self._list.append(dict_)

    @validate_call
    def get_value_from_key(self, key: str) -> Optional[str]:
        if key in self._dict.keys():
            return self._dict[key]

    @validate_call
    def get_key_from_value(self, value: str) -> Optional[str]:
        for dict_key, dict_value in self._dict.items():
            if dict_value == value:
                return dict_key

    @validate_call
    def get_dragon_name(self, id: int) -> Optional[str]:
        key = f"tid_unit_{id}_name"
        return self.get_value_from_key(key)

    @validate_call
    def get_dragon_description(self, id: int) -> Optional[str]:
        key = f"tid_unit_{id}_description"
        return self.get_value_from_key(key)

    @validate_call
    def get_attack_name(self, id: int) -> Optional[str]:
        key = f"tid_attack_name_{id}"
        return self.get_value_from_key(key)

    @validate_call
    def get_skill_name(self, id: int) -> Optional[str]:
        key = f"tid_skill_name_{id}"
        return self.get_value_from_key(key)

    @validate_call
    def get_skill_description(self, id: int) -> Optional[str]:
        key = f"tid_skill_description_{id}"
        return self.get_value_from_key(key)

    @validate_call
    def search_keys(self, query: str) -> list[str]:
        query = (query
            .lower()
            .strip())

        results = []

        for key in self._dict.keys():
            parsed_key = (key
                .lower()
                .strip())

            if query in parsed_key:
                results.append(key)

        return results

    @validate_call
    def search_values(self, query: str) -> list[str] | list:
        query = (query
            .lower()
            .strip())

        results = []

        for value in self._dict.values():
            parsed_value = (value
                .lower()
                .strip())

            if query in parsed_value:
                results.append(value)

        return results

    @validate_call
    def compare(
        self,
        old_localization: Any
    ) -> dict[str, list]:
        if isinstance(old_localization, list):
            old_localization = DictListFilter(old_localization).merge_dicts()

        elif not isinstance(old_localization, dict):
            old_localization = old_localization.dict

        new_fields = []
        edited_fields = []
        deleted_fields = []

        old_localization_keys = old_localization.keys()

        for key in self._dict.keys():
            if key not in old_localization_keys:
                new_fields.append({
                    "key": key,
                    "value": self._dict[key]
                })

        for key in old_localization_keys:
            if key not in self._dict:
                deleted_fields.append({
                    "key": key,
                    "value": old_localization[key]
                })

            elif old_localization[key] != self._dict[key]:
                edited_fields.append({
                    "key": key,
                    "values": {
                        "new": old_localization[key],
                        "old": self._dict[key]
                    }
                })

        return dict(
            new_fields = new_fields,
            edited_fields = edited_fields,
            deleted_fields = deleted_fields
        )

    @property
    def list(self) -> LocalizationList:
        return self._list or []

    @property
    def dict(self) -> LocalizationDict:
        return self._dict or {}


from pydantic import validate_call

from .config import feed_costs

@validate_call
def calculate_feed_cost(
    start_level: int,
    end_level: int
) -> int:
    cost = 0

    for i in range(start_level, end_level):
        cost += feed_costs[i]["total"]

    return cost


from pydantic import validate_call

from ..platform_prefixes import PlatformPrefixes
from ..base import BaseStaticDownloader

class ChestSprite(BaseStaticDownloader):
    @validate_call
    def __init__(
        self,
        image_name: str,
        image_quality: int = 1,
        platform_prefix: str = PlatformPrefixes.IOS
    ) -> None:
        if image_quality == 1:
            image_quality_str = ""

        elif image_quality == 2:
            image_quality_str = "@2x"

        else:
            raise ValueError(f"{image_quality} Not a valid number for image quality of a chest. Choose a number between 1 and 2")
        
        self.url = f"https://{platform_prefix}-static-s1.socialpointgames.com/static/dragoncity/mobile/ui/chests/ui_{image_name}{image_quality_str}.png"